// Glaze Library
// For the license information refer to glaze.hpp

#pragma once

// Convenience header for std::chrono support
// Includes all necessary headers for chrono serialization

#include "glaze/core/chrono.hpp"
#include "glaze/json/chrono_format.hpp"
#include "glaze/json/read.hpp"
#include "glaze/json/write.hpp"
